from collections import OrderedDict
from math import log2, ceil

def get_data_types(data):
	"""
	returns a dictionary with column names and their respective data types
	:param pandas.DataFrame data: a dataframe
	:rtype: dict[str,str]
	"""
	dtypes = data.dtypes
	return OrderedDict(zip(dtypes.index, dtypes.astype(str)))

def get_redshift_data_types(data):
	"""
	returns a dictionary with column names and their respective redshift data types
	:param pandas.DataFrame data: a dataframe
	:rtype: dict[str,str]
	"""
	data_types = get_data_types(data)
	redshift_data_types = OrderedDict()
	for column, data_type in data_types.items():
		if data_type.startswith('int'):
			redshift_data_types[column] = 'INTEGER'
		elif data_type.startswith('float'):
			redshift_data_types[column] = 'REAL'
		elif data_type.startswith('datetime'):
			redshift_data_types[column] = 'TIMESTAMP'
		elif data_type.startswith('bool'):
			redshift_data_types[column] = 'BOOLEAN'
		else:
			max_length = int(max([len(bytes(str(x), 'utf-8')) for x in data[column].values]))
			nearest_power_of_two = 2 ** ceil(log2(max_length+1))-1
			redshift_data_types[column] = f'VARCHAR({nearest_power_of_two})'
	return redshift_data_types

def get_redshift_create_table_query(database, schema, table, data, data_types=None):
	"""
	:param str database: name of database
	:param str schema: name of schema
	:param str table: name of table
	:param pandas.DataFrame data: data to be uploaded
	:param dict[str,str] data_types: a dictionary of the redshift data types
	:rtype: str
	"""
	data_types = data_types or get_redshift_data_types(data=data)
	data_types_str = ', \n'.join([f'"{col}" {dtype}' for col, dtype in data_types.items()])
	return f"""
		CREATE TABLE IF NOT EXISTS {database}.{schema}.{table} (
			{data_types_str}
		)
	"""



from datetime import datetime
import numpy as np


class Snapshot:
	def __init__(self, database):
		"""
		:type database: .Redshift.Redshift
		"""
		database.refresh()
		self._time = datetime.now()
		self._table_data = database.table_data.copy()
		self._column_data = database.column_data.copy()

	@property
	def time(self):
		"""
		:rtype: datetime
		"""
		return self._time

	@property
	def table_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._table_data

	@property
	def column_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._column_data

	def __getstate__(self):
		return {
			'time': self.time,
			'table_data': self.table_data,
			'column_data': self.column_data
		}

	def __setstate__(self, state):
		self._time = state['time']
		self._table_data = state['table_data']
		self._column_data = state['column_data']

	def __sub__(self, other):
		"""
		:type other: Snapshot
		:rtype SnapshotDifference
		"""
		return SnapshotDifference(old_snapshot=other, new_snapshot=self)


class SnapshotDifference:
	def __init__(self, old_snapshot, new_snapshot):
		"""
		:type old_snapshot: Snapshot
		:type new_snapshot: Snapshot
		"""

		column_comparison_cols = ['schema', 'table', 'column']
		columns1 = old_snapshot.column_data[column_comparison_cols]
		columns2 = new_snapshot.column_data[column_comparison_cols]
		self._column_comparisons = columns1.merge(
			right=columns2, on=column_comparison_cols, how='outer', indicator='indicator', validate='one_to_one'
		)
		self._column_comparisons['indicator'].replace(
			to_replace={'left_only': 'missing_column', 'right_only': 'new_column', 'both': 'same_column'},
			inplace=True
		)

		self.new_columns = self._column_comparisons[self._column_comparisons['indicator'] == 'new_column']
		self.missing_columns = self._column_comparisons[self._column_comparisons['indicator'] == 'missing_column']

		#
		#
		#
		#

		table_comp_same = ['schema', 'table']
		table_comparison_different_cols = ['table_id', 'num_rows', 'num_columns', 'mbytes']
		tables1 = old_snapshot.table_data[table_comp_same + table_comparison_different_cols]
		tables2 = new_snapshot.table_data[table_comp_same + table_comparison_different_cols]
		self._table_comparisons = tables1.merge(
			right=tables2, on=table_comp_same, how='outer', indicator='indicator', validate='one_to_one',
			suffixes=['_1', '_2']
		)
		for col in table_comparison_different_cols:

			if col == 'table_id':
				self._table_comparisons[f'{col}_1'] = self._table_comparisons[f'{col}_1'].fillna(-1).astype(int)
				self._table_comparisons[f'{col}_2'] = self._table_comparisons[f'{col}_2'].fillna(-1).astype(int)
				self._table_comparisons[f'{col}_different'] = \
					self._table_comparisons[f'{col}_1'] != self._table_comparisons[f'{col}_2']
			else:
				self._table_comparisons[f'{col}_1'] = self._table_comparisons[f'{col}_1'].fillna(0).astype(int)
				self._table_comparisons[f'{col}_2'] = self._table_comparisons[f'{col}_2'].fillna(0).astype(int)
				self._table_comparisons[f'{col}_difference'] = \
					self._table_comparisons[f'{col}_2'] - self._table_comparisons[f'{col}_1']

		self._table_comparisons['indicator'].replace(
			to_replace={'left_only': 'missing_table', 'right_only': 'new_table', 'both': 'same_table'}, inplace=True
		)

		self.table_growth = self._table_comparisons[
			(self._table_comparisons['indicator'] == 'same_table') & (
				(self._table_comparisons['num_rows_difference'] != 0) |
				(self._table_comparisons['num_columns_difference'] != 0) |
				(self._table_comparisons['mbytes_difference'] != 0) |
				(self._table_comparisons['table_id_different'])
			)
		]

		column_changes = self._column_comparisons.copy()
		column_changes['num_new_columns'] = (column_changes['indicator'] == 'new_column') * 1
		column_changes['num_missing_columns'] = (column_changes['indicator'] == 'missing_column') * 1
		column_changes = column_changes[['schema', 'table', 'num_new_columns', 'num_missing_columns']].groupby(
			['schema', 'table']
		).sum().reset_index(drop=False)
		table_changes = self._table_comparisons[['schema', 'table', 'indicator']]
		table_changes = table_changes.merge(
			right=column_changes, on=['schema', 'table'], how='outer', validate='one_to_one'
		)

		table_changes['indicator'] = np.where(
			(
				(table_changes['num_new_columns'] > 0) | (table_changes['num_new_columns'] > 0)
			) & (table_changes['indicator'] == 'same_table'),
			'table_changed',
			table_changes['indicator']
		)
		self.table_changes = table_changes








from .Column import Column
from .Metadata import Metadata
import warnings


class Table:
	def __init__(self, name, schema, echo=None):
		"""
		:type name: str
		:type schema: Schema
		:type echo: int or NoneType
		"""
		self._name = name
		self._schema = schema
		self._data = None
		self._columns_info = None
		self._columns = None
		self._dictionary = None
		self._metadata = None
		if name not in schema.table_names:
			raise KeyError(f'"{name}" not in "{schema}"')
		if echo is None:
			self._echo = max(0, self.schema.echo-1)
		else:
			self._echo = echo

	def __getstate__(self):
		return {
			'name': self.name,
			'columns_info': self.column_info,
			'columns': self._columns,
			'dictionary': self.dictionary,
			'metadata': self._metadata,
			'echo': self._echo
		}

	def __setstate__(self, state):
		self._name = state['name']
		self._schema = None
		self._columns_info = state['columns_info']
		self._columns = state['columns']
		self._dictionary = state['dictionary']
		self._metadata = state['metadata']
		self._echo = state['echo']
		self._data = None
		self.add_table_to_columns()

	def reset(self):
		self._data = None
		self._columns_info = None
		if self._columns is not None:
			for column in self.columns:
				column.reset()
		self._dictionary = None

	def __eq__(self, other):
		"""
		:type other: Table
		:rtype: bool
		"""
		return self.column_names == other.column_names

	def __ne__(self, other):
		return not self.__eq__(other=other)

	def __sub__(self, other):
		"""
		:type other: Table
		:rtype:
		"""
		if not isinstance(other, self.__class__):
			raise TypeError(f'{self} is of type {type(self)} but {other} is of type {type(other)}!')
		if self.schema != other.schema:
			warnings.warn(
				f'The two tables belong to different schemas: "{self.schema}" and "{other.schema}" respectively!'
			)
		if self.name != other.name:
			warnings.warn(
				f'The two tables have different names: "{self.name}" and "{other.name}" respectively!'
			)
		return {
			'num_rows_difference': self.num_rows - other.num_rows,
			'missing_column_names': [x for x in other.column_names if x not in self.column_names],
			'new_column_names': [x for x in self.column_names if x not in other.column_names]
		}

	@property
	def metadata(self):
		"""
		:rtype: Metadata
		"""
		return self._metadata

	@metadata.setter
	def metadata(self, metadata):
		"""
		:type metadata: dict or Metadata
		"""
		if self._metadata is None:
			if not isinstance(metadata, Metadata):
				metadata = Metadata(**metadata)
			self._metadata = metadata
		else:
			self._metadata.update(metadata, inplace=True)

	@property
	def all_metadata(self):
		"""
		:rtype: Metadata
		"""
		return self.schema.metadata.update(self.metadata, inplace=False)

	def add_metadata(self, metadata, column=None):
		"""
		:type metadata: dict or Metadata
		:param str column: name of column
		"""
		if column is None:
			self.metadata = metadata
		else:
			self.column[column].metadata = metadata

	@property
	def echo(self):
		return self._echo is not None and self._echo

	@echo.setter
	def echo(self, echo):
		self._echo = echo

	@property
	def schema(self):
		return self._schema

	@property
	def name(self):
		return self._name

	@property
	def data(self):
		if self._data is None:
			query = 'SELECT * FROM ' + self.schema.name + '.' + self.name
			self._data = self.schema.database.get_dataframe(query=query, echo=self.echo)
		return self._data

	def get_head(self, num_rows=5):
		query = f'SELECT TOP {num_rows} * FROM ' + self.schema.name + '.' + self.name
		return self.schema.database.get_dataframe(query=query, echo=self.echo)

	@property
	def column_info(self):
		if self._columns_info is None:
			columns_data = self.schema.database.column_data
			self._columns_info = columns_data[
				(columns_data['schema'] == self.schema.name) & (columns_data['table'] == self.name)
			].copy().reset_index(drop=True)
		return self._columns_info

	@property
	def column_names(self):
		"""
		:rtype: list of str
		"""
		return list(self.column_info['column'].values)

	@property
	def column(self):
		"""
		:rtype: dict[str,Column]
		"""
		if self._columns is None:
			self._columns = {column_name: Column(name=column_name, table=self) for column_name in self.column_names}
		return self._columns.copy()

	@property
	def columns(self):
		"""
		:rtype: list[Column]
		"""
		return [column for column in self.column.values()]

	def add_table_to_columns(self):
		if self._columns is not None:
			for column in self.columns:
				column._table = self

	@property
	def shape(self):
		database_shape = self.schema.database.shape
		return database_shape[(database_shape['table'] == self.name) & (database_shape['schema'] == self.schema.name)]

	@property
	def dictionary(self):
		"""
		:rtype: dict
		"""
		if self._dictionary is None:
			self._dictionary = self.shape.iloc[0].to_dict()
		return self._dictionary.copy()

	@property
	def num_rows(self):
		return self.dictionary['num_rows']

	@property
	def num_columns(self):
		return self.dictionary['num_columns']

	def __str__(self):
		return f'{str(self.schema)}.{self.name}'

	def __repr__(self):
		return str(self)

	def __getitem__(self, item):
		try:
			return self.column[item]
		except KeyError:
			result = self.all_metadata[item]
			if result is None:
				raise AttributeError(f"table '{self.name}' has no attribute, column, or metadata '{item}'")
			else:
				return result

	def __getattr__(self, item):
		return self[item]



from s3fs import S3FileSystem
from pandas import read_csv
from pickle import dump as pickle_dump
from pickle import load as pickle_load
from psycopg2 import connect as psycopg2_connect
from csv import QUOTE_NONNUMERIC
from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame as SparkDF
from .S3File import S3Files


class S3:
	def __init__(self, key=None, secret=None, iam_role=None, root='s3://', spark=None):
		"""
		starts an S3 connection
		:type key: str or NoneType
		:type secret: str or NoneType
		:type iam_role: str or NoneType
		:type root: str or NoneType
		:type spark: SparkSession or NoneType
		"""
		self._key = key
		self._secret = secret
		self._iam_role = iam_role
		if root is None:
			root = ''
		self._root = root
		self._spark = spark

		self.list = self.ls
		self.move = self.mv
		self.dir = self.ls
		self.copy = self.cp
		self.delete = self.rm

	@property
	def file_system(self):
		return S3FileSystem(key=self._key, secret=self._secret, use_ssl=False)

	def _get_absolute_path(self, path):
		if not path.startswith(self._root):
			path = self._root + path
		return path

	def ls(self, path, exclude_empty=False, detail=False, sort_by='path', sort_reverse=True, **kwargs):
		path = self._get_absolute_path(path)
		files = self.file_system.ls(path=path, detail=detail, **kwargs)

		# if the path is only a file ls will return itself.
		if len(files) == 1:
			if self._root + files[0] == path or files[0] == path:
				return []

		if exclude_empty:
			files = [file for file in files if self.get_size(file) > 0]
		if detail:
			files = S3Files(files)
			files.sort(by=sort_by, reverse=sort_reverse)
		else:
			if sort_by is not None:
				files.sort()
		return files

	def mv(self, path1, path2, recursive=True, max_depth=None, **kwargs):
		return self.file_system.mv(path1=path1, path2=path2, recursive=recursive, maxdepth=max_depth, **kwargs)

	def cp(self, path1, path2, recursive=True, on_error=None, **kwargs):
		return self.file_system.copy(path1=path1, path2=path2, recursive=recursive, on_error=on_error, **kwargs)

	def rm(self, path, recursive=True, **kwargs):
		path = self._get_absolute_path(path)
		return self.file_system.delete(path=path, recursive=recursive, **kwargs)

	def mkdir(self, path, **kwargs):
		path = self._get_absolute_path(path)
		return self.file_system.mkdir(path=path, **kwargs)

	def tree(self, path, depth_limit=None, indentation='\t'):
		def _get_tree(_path, _depth):
			subs = self.ls(_path)
			if _depth == 0:
				name = _path
			else:
				name = self.get_file_name_and_extension(_path)

			if len(subs) == 0:
				return f'{indentation * _depth}{name}'
			else:
				if depth_limit is None:
					return f'{indentation * _depth}{name}/\n' + '\n'.join(
						[_get_tree(_path=sub, _depth=_depth + 1) for sub in subs]
					)
				elif depth_limit > _depth:
					return f'{indentation * _depth}{name}/\n' + '\n'.join(
						[_get_tree(_path=sub, _depth=_depth + 1) for sub in subs]
					)
				else:
					return f'{indentation * _depth}{name}/\n{indentation * (_depth + 1)}...'
		print(_get_tree(_path=path, _depth=0))

	def exists(self, path):
		path = self._get_absolute_path(path)
		return self.file_system.exists(path=path)

	def write_bytes(self, path, bytes):
		path = self._get_absolute_path(path)
		with self.file_system.open(path=path, mode='wb') as f:
			f.write(bytes)

	def get_size(self, path):
		path = self._get_absolute_path(path)
		return self.file_system.size(path=path)

	def write_csv(self, data, path, index=False, encoding='utf-8', **kwargs):
		"""

		:type data: SparkDF
		:param path:
		:param index:
		:param kwargs:
		:return:
		"""
		path = self._get_absolute_path(path)
		bytes = data.to_csv(path_or_buf=None, quoting=QUOTE_NONNUMERIC, index=index, **kwargs).encode(encoding)
		self.write(path=path, bytes=bytes)

	def read_bytes(self, path):
		with self.file_system.open(path=self._root+path, mode='rb') as f:
			result = f.read()
		return result

	def read_csv(self, path, encoding='utf-8', **kwargs):
		path = self._get_absolute_path(path)
		with self.file_system.open(self._root+path, 'rb', ) as f:
			df = read_csv(f, encoding=encoding, **kwargs)
		return df

	def write_pickle(self, obj, path, mode='overwite'):
		"""

		:type obj: SparkDF
		:type path: str
		:type mode: str
		:rtype: bool
		"""
		path = self._get_absolute_path(path)
		if self.exists(path=path):
			if mode == 'overwrite':
				self.delete(path=path)
			else:
				raise FileExistsError(f'File "{path}" exists on S3!')

		with self.file_system.open(path=path, mode='wb') as f:
			try:
				return obj.to_pickle(f)
			except:
				if self.exists(path=path):
					self.delete(path=path)
				try:
					return pickle_dump(obj=obj, file=f)
				except Exception as e:
					if self.exists(path=path):
						self.delete(path=path)
					raise e

	def read_pickle(self, path):
		path = self._get_absolute_path(path)
		with self.file_system.open(path=path, mode='rb') as f:
			obj = pickle_load(file=f)
		return obj

	def copy_to_redshift(self, path, redshift, schema, table, truncate=False, create_table=False):
		path = self._get_absolute_path(path)
		if create_table:
			data = self.read_csv(path=path)
			redshift.create_table(data=data, name=table, schema=schema)

		connection = psycopg2_connect(f"""
			dbname='{redshift._database}' port='{redshift._port}' 
			user='{redshift._user_id}' password='{redshift._password}' 
			host='{redshift._server}'
		""")

		cursor = connection.cursor()

		if truncate:
			cursor.execute(f"TRUNCATE TABLE {schema}.{table}")

		if self._iam_role:
			credentials = f"IAM_ROLE '{self._iam_role}'"
		else:
			credentials = f"CREDENTIALS 'aws_access_key_id={self._key};aws_secret_access_key={self._secret}'"

		cursor.execute(f"""
			COPY {schema}.{table} FROM '{self._root+path}' 
			{credentials}
			FORMAT AS CSV ACCEPTINVCHARS EMPTYASNULL IGNOREHEADER 1;commit;
		""")

		connection.close()

	def write_parquet(self, data, path, mode='overwrite'):
		"""
		saves a Spark DataFrame to a path on S3 and returns the list of parquet files
		:type data: SparkDF
		:type path: str
		:type mode: str
		:rtype: list[str]
		"""
		path = self._get_absolute_path(path)

		if mode == 'overwrite' and self.exists(path=path):
			self.rm(path=path, recursive=True)

		data.write.mode(mode).save(path=path)
		return self.ls(path=path)

	@property
	def spark(self):
		"""
		:rtype: SparkSession
		"""
		if self._spark is None:
			return RuntimeError('S3 does not have access to a SparkSession')
		return self._spark

	@staticmethod
	def get_file_name_and_extension(path):
		"""
		returns a file name from a path
		:type path: str
		:rtype: str
		"""
		return path.strip('/').split('/')[-1]

	@classmethod
	def get_file_name(cls, path):
		"""
		returns a file name from a path
		:type path: str
		:rtype: str
		"""
		return cls.get_file_name_and_extension(path=path).split('.')[0]

	@classmethod
	def get_file_extension(cls, path):
		"""
		returns a file extension from a path
		:type path: str
		:rtype: str
		"""
		name_and_extension = cls.get_file_name_and_extension(path=path).split('.')
		if len(name_and_extension) > 1:
			return '.'.join(name_and_extension[1:])
		else:
			return None

	def is_file(self, path):
		return self.file_system.isfile(path=path)

	def is_dir(self, path):
		return self.file_system.isdir(path)

	@property
	def json(self):
		return self.file_system.to_json()

	def is_parquet_file(self, path):
		"""
		returns True if a path is a parquet file
		:type path: str
		:rtype: bool
		"""
		if self.get_size(path=path) == 0:
			return False
		else:
			n_and_e = self.get_file_name_and_extension(path=path).lower()
			return n_and_e.startswith('part-') and n_and_e.endswith('.parquet')

	def read_parquet(self, path, spark=None, parallel=True):
		"""
		reads parquet files inside a path and returns the data
		:type path: str
		:type spark: SparkSession or NoneType
		:type parallel: bool
		:rtype: SparkDF
		"""
		if spark is None:
			spark = self.spark

		files = self.ls(path=path, exclude_empty=True)
		if len(files) == 1:
			file = files[0]
			if self.get_size(file) > 0:
				if not self.is_parquet_file(path=file):
					print(f'"{file}" does not appear to be a parquet file!')
				return spark.read.load(file)
			else:
				raise FileNotFoundError(f'"{file}" is empty!')

		elif parallel:
			return spark.read.load(path=f'{path}/part-*.parquet')

		else:
			parquet_files = [file for file in files if self.is_parquet_file(path=file)]
			result = None
			for parquet in parquet_files:
				data = spark.read.load(self._get_absolute_path(parquet))
				if result is None:
					result = data
				else:
					result = result.union(data)

			return result

	load_parquet = read_parquet
	save_parquet = write_parquet
	load_pickle = read_pickle
	save_pickle = write_pickle

	def save(self, obj, path, mode='overwrite'):
		"""
		writes an object to path as a parquet if the object is a Spark DataFrame or a pickle otherwise
		:type obj: object
		:type path: str
		:type mode: str
		:rtype: bool
		"""
		if isinstance(obj, SparkDF):
			if path.endswith('.pickle'):
				raise ValueError(f'Spark DataFrame cannot be saved into a pickle file: "{path}"')

			if not path.endswith('.parquet'):
				path += '.parquet'

			return self.write_parquet(data=obj, path=path, mode=mode)

		else:
			if path.endswith('.parquet'):
				raise ValueError(f'{type(path)} cannot be saved into a parquet file: "{path}"')

			if not path.endswith('.pickle'):
				path += '.pickle'

			return self.write_pickle(obj=obj, path=path, mode=mode)

	def load(self, path, spark=None):
		"""
		reads an object from a path. If the extension is .parquet then the object is assumed to be a Spark DataFrame
		and the read_parquet method is used, otherwise a pickle will be read.
		:type path: str
		:type spark: SparkSession
		:rtype: bool
		"""
		if path.endswith('.parquet'):
			return self.read_parquet(path=path, spark=spark or self.spark)

		elif path.endswith('.pickle'):
			return self.read_pickle(path=path)

		else:
			raise NotImplementedError(f'load is not implemented for {path}!')



class Metadata:
	def __init__(self, **kwargs):
		self._dictionary = {
			key:value for key, value in kwargs.items()
			if value is not None and value != '' and value != {} and value != []
		}

	def copy(self):
		return Metadata(**self._dictionary)

	def __getstate__(self):
		return self._dictionary

	def __setstate__(self, state):
		self._dictionary = state

	def __getitem__(self, item):
		if item in self._dictionary:
			return self._dictionary[item]
		else:
			return None

	def __setitem__(self, key, value):
		self._dictionary[key] = value

	def __contains__(self, item):
		return item in self._dictionary

	def update(self, x, inplace=False):
		"""
		:type x: dict or Metadata
		:rtype: Metadata
		"""
		if isinstance(x, Metadata):
			x = x._dictionary
		if inplace:
			metadata = self
		else:
			metadata = self.copy()

		metadata._dictionary.update({
			key: value for key, value in x.items() if value is not None and value != '' and value != {} and value != []
		})
		return metadata

	def __repr__(self):
		return repr(self._dictionary)

	def __str__(self):
		return str(self._dictionary)


from sqlalchemy import create_engine, text
import time
from numpy import dtype as numpy_dtype
from pandas import read_sql_query, concat, DataFrame
import psycopg2
from .get_redshift_create_table_query import get_redshift_create_table_query


class BasicRedshift:
	def __init__(self, user_id, password, server, database, port='5439'):
		"""
		:type server: str
		:type port: str
		:type database: str
		:type user_id: str
		:type password: str
		"""
		self._server = server
		self._port = port
		self._database = database
		self._user_id = user_id
		self._password = password
		self._engine = create_engine(self._engine_string)

	def __getstate__(self):
		return {
			'server': self._server,
			'port': self._port,
			'database': self._database,
			'user_id': self._user_id,
			'password': self._password
		}

	def __setstate__(self, state):
		self._server = state['server']
		self._port = state['port']
		self._database = state['database']
		self._user_id = state['user_id']
		self._password = state['password']
		self._engine = create_engine(self._engine_string)

	@property
	def _engine_string(self):
		return f'postgresql://{self._user_id}:{self._password}@{self._server}:{self._port}/{self._database}'

	@property
	def name(self):
		return self._database

	def __str__(self):
		return f'{self._server}/{self.name}'

	def run(self, query):
		connection = None
		try:
			connection = psycopg2.connect(f"""
				dbname='{self._database}' port='{self._port}' 
				user='{self._user_id}' password='{self._password}' 
				host='{self._server}'
			""")
			cursor = connection.cursor()
			cursor.execute(query)
			connection.commit()
		except (Exception, psycopg2.DatabaseError) as error:
			print(error)
		finally:
			if connection is not None:
				connection.close()

	def change_password(self, new_password):
		"""
		:type new_password: str
		:return:
		"""
		the_query = "ALTER USER " + self._user_id + " PASSWORD " + "'" + new_password + "'"
		connection = self._engine.connect()
		result = connection.execute(text(the_query))
		self._password = new_password
		return result

	def get_dataframe(self, query, echo=1):
		"""
		:type query: str
		:type echo: int
		:return:
		"""
		start_time = time.time()
		if echo:
			print('\n', query, '\n', sep='')

		result = read_sql_query(query, self._engine)
		elapsed_time = time.time() - start_time
		if elapsed_time >= 3600:
			elapsed_time = elapsed_time / 3600
			time_unit = 'hours'
		elif elapsed_time >= 60:
			elapsed_time = elapsed_time / 60
			time_unit = 'minutes'
		else:
			time_unit = 'seconds'

		if echo:
			print(f'shape:{result.shape}  elapsed time:{round(elapsed_time, 2)}{time_unit}')

		return result


	# schemas:
	_schemas_query = "select * from pg_namespace where nspowner<>'1';"

	def get_schemas(self, echo=1):
		"""
		:type echo: int
		:return:
		"""
		the_query = self._schemas_query
		return self.get_dataframe(query=the_query, echo=echo).sort_values('nspname')

	# tables:
	def _get_tables_data_depricated(self, echo=1):
		the_query = """
			SELECT table_schema AS "schema", table_name AS "table", table_type AS "type" 
			FROM SVV_TABLES WHERE table_schema NOT IN ('admin', 'information_schema', 'metadata', 'public') 
			AND table_schema NOT LIKE 'pg_%%' AND table_type IN ('BASE TABLE', 'VIEW') 
			ORDER BY table_schema, table_name
		"""

		return self.get_dataframe(query=the_query, echo=echo)

	# counts the number of cases with similar values for columns
	# it's good for finding non-unique cases
	def count_similar_rows(self, schema, table, columns, echo=1):
		"""
		:type schema: str
		:type table: str
		:type columns: list of str
		:type echo: int
		:rtype: DataFrame
		"""
		columns_query = '"' + '", "'.join(columns) + '"'

		query = 'SELECT ' + columns_query + ', COUNT(*) AS n '
		query += 'FROM ' + schema + '.' + table + ' GROUP BY ' + columns_query + ' ORDER BY n DESC;'
		return self.get_dataframe(query, echo=echo)

	# close?
	def close(self):
		self._engine.dispose()

	def get_table_scrape_dates_query(self, schema, table, period='month', scrape_date_column='scrape_date'):
		"""
		:type schema: str
		:type table: str
		:type period: str
		:type scrape_date_column: str
		:rtype: str
		"""

		if period == 'year':
			the_query = (
					'SELECT "schema", "table", "year", COUNT(*) AS rows FROM --\n'
					'( --\n'
					'    SELECT --\n'
					'        \'' + schema + '\' AS "schema", --\n'
					'        \'' + table + '\' AS "table", --\n'
					'        DATEPART(year, "' + scrape_date_column + '") AS "year", --\n'
					'        DATEPART(month, "' + scrape_date_column + '") AS "month" --\n'
					'    FROM ' + self._database + '.' + schema + '.' + table + ' --\n'
					') X GROUP BY "schema", "table", "year"'
			)
		elif period == 'month':
			the_query = (
					'SELECT "schema", "table", "year", "month", COUNT(*) AS rows FROM --\n'
					'( --\n'
					'    SELECT --\n'
					'        \'' + schema + '\' AS "schema", --\n'
					'        \'' + table + '\' AS "table", --\n'
					'        DATEPART(year, "' + scrape_date_column + '") AS "year", --\n'
					'        DATEPART(month, "' + scrape_date_column + '") AS "month" --\n'
					'    FROM ' + self._database + '.' + schema + '.' + table + ' --\n'
					') X GROUP BY "schema", "table", "year", "month"'
			)
		else:
			the_query = (
					'SELECT "schema", "table", "date", COUNT(*) AS rows FROM --\n'
					'( --\n'
					'    SELECT --\n'
					'        \'' + schema + '\' AS "schema", --\n'
					'        \'' + table + '\' AS "table", --\n'
					'        TRUNC("' + scrape_date_column + '") AS "date" --\n' 
					'    FROM ' + self._database + '.' + schema + '.' + table + ' --\n'
					') X GROUP BY "schema", "table", "date" '
			)
		return the_query

	def get_table_scrape_dates(self, schema, table, period='month', scrape_date_column='scrape_date', echo=1):
		"""
		:type schema: str
		:type table: str
		:type period: str
		:type scrape_date_column: str
		:type echo: int
		:rtype: DataFrame
		"""
		the_query = self.get_table_scrape_dates_query(
			schema=schema, table=table, period=period,
			scrape_date_column=scrape_date_column
		)
		return self.get_dataframe(query=the_query, echo=echo)

	def get_scrape_dates(self, schema=None, period='month', scrape_date_column='scrape_date', echo=1):
		"""
		:type schema: str
		:type period: str
		:type scrape_date_column: str
		:type echo: int
		:rtype: DataFrame
		"""
		tables_columns = self.get_columns_data(schema=schema, echo=echo)

		scrape_date_tables = tables_columns[tables_columns['column'] == scrape_date_column]

		the_queries = scrape_date_tables.apply(
			lambda x: self.get_table_scrape_dates_query(
				schema=x['schema'], table=x['table'], period=period,
				scrape_date_column=scrape_date_column
			),
			axis=1
		)

		union_query = " UNION --\n".join(the_queries)

		if period == 'year':
			order_query = ' ORDER BY "schema", "table", "year", "month" '
		elif period == 'month':
			order_query = ' ORDER BY "schema", "table", "year" '
		else:
			order_query = ' ORDER BY "schema", "table", "date" '

		return self.get_dataframe(query=union_query + '--\n' + order_query, echo=echo)

	def create_table2(self, data, name, schema, index=False, if_exists='replace'):
		"""
		:type data: DataFrame
		:type name: str
		:type schema: str
		:type index:
		:type if_exists: str
		"""
		null_data = data.iloc[[0], :]
		max_data = data.iloc[[0], :]
		for col in data.columns:
			if any(data[col].isnull()):
				null_data[col] = None
			if data[col].dtype in [numpy_dtype('float64'), numpy_dtype('int64')]:
				max_data[col] = data[col].max()
			else:
				max_string_length = int(data[col].str.len().max())
				max_data[col] = 'x'*(max_string_length+1)
		temp_data = concat([null_data, max_data])
		print(temp_data)
		temp_data.to_sql(name=name, schema=schema, con=self._engine, index=index, if_exists=if_exists)

	def create_table(self, data, name, schema):
		query = get_redshift_create_table_query(database=self.name, schema=schema, table=name, data=data)
		self.run(query=query)
		return query

	def get_errors(self, limit=10, echo=1):
		"""
		:type limit: int
		:type echo: int
		:rtype: DataFrame
		"""
		the_query = (
			"SELECT --\n"
			"	starttime, err_reason, filename, colname, type, --\n"
			"	line_number, position, col_length, raw_field_value, raw_line --\n"
			"FROM stl_load_errors order by starttime desc LIMIT {};"
		).format(limit)

		return self.get_dataframe(query=the_query, echo=echo)

	def get_columns_data(self, schema=None, echo=1):
		"""
		:type schema: str
		:type echo: int
		:rtype: DataFrame
		"""

		the_query = self.columns_query
		order_query = 'ORDER BY datname, nspname, relname, attname;'

		if schema is not None:
			the_query += 'AND TRIM(pg_namespace.nspname) = \'' + schema + '\' '

		the_query += order_query

		result = self.get_dataframe(the_query, echo=echo)
		return result[~result['table'].str.startswith('#')]

	@property
	def columns_query(self):
		_columns_query = (
			'SELECT DISTINCT --\n'
			'	TRIM(pg_database.datname) AS "database",  --\n'
			'	TRIM(pg_namespace.nspname) AS "schema",  --\n'
			'	TRIM(pg_class.relname) AS "table",  --\n'
			'	TRIM(pg_attribute.attname) AS "column" --\n'
			'FROM pg_class '
			'LEFT JOIN pg_namespace ON pg_namespace.oid = pg_class.relnamespace --\n'
			'LEFT JOIN stv_tbl_perm ON pg_class.oid = stv_tbl_perm.id --\n'
			'LEFT JOIN pg_database ON pg_database.oid = stv_tbl_perm.db_id --\n'
			'LEFT JOIN pg_attribute ON pg_attribute.attrelid = stv_tbl_perm.id --\n'
			'WHERE TRIM(pg_database.datname) = \'' + self._database + '\' '
		)
		return _columns_query

	def get_tables_data_query(self, schema=None):
		"""
		:type schema: str
		:rtype: str
		"""
		if schema is not None:
			where_clause = f'WHERE "schema" = \'{schema}\' '
		else:
			where_clause = ''

		the_query = f"""
			SELECT 
				X.id AS table_id, 
				X."database", 
				X."schema", 
				X."table", 
				X."rows" AS "num_rows", 
				COUNT(DISTINCT pg_attribute.attname) AS "num_columns", 
				ISNULL(size_table.mbytes, 0) AS mbytes 
			FROM 
			( 
				SELECT  
					stv_tbl_perm.id, 
					TRIM(pg_database.datname) AS "database", 
					TRIM(pg_namespace.nspname) AS "schema", 
					TRIM(pg_class.relname) AS "table", 
					SUM(stv_tbl_perm.rows) AS "rows" 
				FROM pg_class 
				LEFT JOIN pg_namespace ON pg_namespace.oid = pg_class.relnamespace 
				LEFT JOIN stv_tbl_perm ON pg_class.oid = stv_tbl_perm.id 
				LEFT JOIN pg_database ON pg_database.oid = stv_tbl_perm.db_id 
				WHERE TRIM(pg_database.datname) =\'{self._database}\' GROUP BY id, datname, nspname, relname 
			) X 
			LEFT JOIN 
			( 
				SELECT tbl as id, COUNT(*) AS mbytes FROM stv_blocklist GROUP BY tbl 
			) size_table ON size_table.id = X.id 
			LEFT JOIN pg_attribute ON pg_attribute.attrelid = X.id 
			{where_clause} 
			GROUP BY table_id, "database", "schema", "table", "rows", mbytes 
			ORDER BY "database", "schema", "table", "rows"; 

		"""
		return the_query

	def get_tables_data(self, schema=None, echo=0):
		result = self.get_dataframe(query=self.get_tables_data_query(schema=schema), echo=echo)
		return result[~result['table'].str.startswith('#')]



from .BasicRedshift import BasicRedshift
from .Schema import Schema
from .Table import Table
from .Column import Column
from .Snapshot import Snapshot


class Redshift(BasicRedshift):
	def __init__(self, user_id, password, server, database, port='5439', echo=0):
		super().__init__(user_id=user_id, password=password, port=port, server=server, database=database)
		self._schema_dict = None
		self._hierarchy = None
		self._table_data = None
		self._column_data = None
		self._echo = echo

	def reset(self):
		if self._schema_dict is not None:
			for schema in self.schemas:
				schema.reset()
		self._hierarchy = None
		self._table_data = None
		self._column_data = None

	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			'schemas': self._schema_dict,
			'hierarchy': self._hierarchy,
			'table_data': self._table_data,
			'column_data': self._column_data,
			'echo': self._echo,
		})
		return state

	def __setstate__(self, state):
		super().__setstate__(state)
		self._schema_dict = state['schemas']
		self._hierarchy = state['hierarchy']
		self._table_data = state['table_data']
		self._column_data = state['column_data']
		self._echo = state['echo']
		self.add_database_to_schemas()

	@property
	def echo(self):
		return self._echo is not None and self._echo

	@echo.setter
	def echo(self, echo):
		self._echo = echo

	def _update_tables_data(self):
		self._table_data = self.get_tables_data(echo=self.echo)

	@property
	def table_data(self):
		"""
		:rtype: DataFrame
		"""
		if self._table_data is None:
			self._update_tables_data()
		return self._table_data

	shape = table_data

	def _update_columns_data(self):
		self._column_data = self.get_columns_data(echo=self.echo)

	@property
	def column_data(self):
		"""
		:rtype: DataFrame
		"""
		if self._column_data is None:
			self._update_columns_data()
		return self._column_data

	def _update_hierarchy(self):
		self._hierarchy = {
			schema: list(data['table'].unique()) for schema, data in self.table_data.groupby(by='schema')
		}

	@property
	def hierarchy(self):
		"""
		:rtype: dict[str,list[str]]
		"""
		if self._hierarchy is None:
			self._update_hierarchy()
		return self._hierarchy

	def _update_schemas(self):
		self._schema_dict = {schema: Schema(name=schema, database=self) for schema in self.get_schema_list()}

	@property
	def schema(self):
		"""
		:rtype: dict[str,Schema]
		"""
		if self._schema_dict is None:
			self._update_schemas()
		return self._schema_dict

	@property
	def schemas(self):
		"""
		:rtype: list[Schema]
		"""
		return [schema for schema in self.schema.values()]

	@property
	def table_list(self):
		"""
		:rtype: list[Table]
		"""
		return [table for schema in self.schemas for table in schema.tables]

	@property
	def columns(self):
		"""
		:rtype: list[Column]
		"""
		return [column for table in self.table_list for column in table.columns]

	def add_database_to_schemas(self):
		if self._schema_dict is not None:
			for schema in self.schemas:
				schema._database = self
				schema.add_schema_to_tables()

	def get_schema_list(self):
		return list(self.table_data['schema'].unique())

	def get_table(self, schema, table):
		return self.schema[schema].table[table]

	def __getitem__(self, item):
		return self.schema[item]

	def __getattr__(self, item):
		return self[item]

	def refresh(self):
		self.reset()
		self._update_tables_data()
		self._update_columns_data()
		self._update_hierarchy()
		self._update_schemas()

	def take_snapshot(self):
		return Snapshot(database=self)

	def add_metadata(self, metadata, schema, table=None, column=None):
		"""
		:type metadata: dict or Metadata
		:param str schema: name of schema
		:param str table: name of table
		:param str column: name of column
		"""
		if table == '':
			table = None
		if column == '':
			column = None

		self.schema[schema].add_metadata(metadata=metadata, table=table, column=column)
